 #include <rewindow/types/window.h>
#include "rewindow/logger/logger.h"
std::string RWindowFlagToStr(RWindowFlags flag) {
    switch (flag) {
        case RWindowFlags::IN_FOCUS: return "IN_FOCUS";
        case RWindowFlags::FULLSCREEN: return "FULLSCREEN";
        case RWindowFlags::RESIZABLE: return "RESIZEABLE";
        case RWindowFlags::VSYNC: return "VSYNC";
        case RWindowFlags::QUIT: return "QUIT";
        case RWindowFlags::MAX_FLAG: return "MAX_FLAG";
        default:
            return "unimplemented flag";
    }
};

using namespace ReWindow;


static RWindow* extant;

RWindow::RWindow(const std::string& wTitle, int wWidth, int wHeight, RenderingAPI wRenderer, bool wFullscreen, bool wResizable, bool wVsync)
: title(wTitle), width(wWidth), height(wHeight), renderer(wRenderer), fullscreen_mode(wFullscreen), resizable(wResizable), vsync(wVsync),
flags{false,wFullscreen,wResizable,wVsync} {
    extant = this;
}

RWindow::~RWindow() {
    if (open)
        DestroyOSWindowHandle();
}



Vector2 RWindow::GetMouseCoordinates() const {
    return currentMouse.Position;
}

bool RWindow::GetFlag(RWindowFlags flag) const {
    return flags[(int) flag];
}

bool RWindow::IsAlive() const {
    return (!closing) && open;
}

void RWindow::SetFullscreen(bool fs) {
    if (fs)
        Fullscreen();
    else
        RestoreFromFullscreen();
}

#pragma region Event Processors Implementation
void RWindow::processFocusIn()
{
    RWindowEvent event {};
    OnFocusGain(event);
    OnFocusGainEvent(event);
    LogEvent(event);
}

void RWindow::processFocusOut()
{
    RWindowEvent event {};
    OnFocusLost(event);
    OnFocusLostEvent(event);
    LogEvent(event);
}

void RWindow::processMousePress(const MouseButton& btn)
{
    currentMouse.Set(btn, true);
    auto event = MouseButtonDownEvent(btn);
    OnMouseButtonDown(event);
    OnMouseButtonDownEvent(event);
    Input::OnMouseButtonEvent(MouseButtonEvent(btn, true));
    Input::OnMouseDown(event);
    LogEvent(event);
}

void RWindow::processMouseMove(Vector2 last_pos, Vector2 new_pos)
{
    currentMouse.Position = new_pos;
    auto event = MouseMoveEvent(new_pos);
    OnMouseMove(event);
    OnMouseMoveEvent(event);
    Input::OnMouseMove(event);
    LogEvent(event);
}

void RWindow::processMouseRelease(const MouseButton& btn)
{
    currentMouse.Set(btn, false);
    auto event = MouseButtonUpEvent(btn);
    OnMouseButtonUp(event);
    OnMouseButtonUpEvent(event);
    Input::OnMouseButtonEvent(MouseButtonEvent(btn, false));
    Input::OnMouseUp(event);
    LogEvent(event);

}

void RWindow::processKeyRelease(Key key) {
    currentKeyboard.PressedKeys[key] = false;
    auto event = KeyUpEvent(key);
    OnKeyUp(event);
    OnKeyUpEvent(event);
    Input::OnKeyboardEvent(KeyboardEvent(key, KeyState::Released));
    Input::OnKeyEvent(KeyboardEvent(key, KeyState::Released));
    Input::OnKeyUp(event);
    LogEvent(event);
}

void RWindow::processKeyPress(Key key) {
    currentKeyboard.PressedKeys[key] = true;
    auto event = KeyDownEvent(key);
    OnKeyDown(event);
    OnKeyDownEvent(event);
    Input::OnKeyDown(event);
    Input::OnKeyboardEvent(KeyboardEvent(key, KeyState::Pressed));
    Input::OnKeyEvent(KeyboardEvent(key, KeyState::Pressed));
    LogEvent(event);
}

void RWindow::processOnClose()
{
    auto event = RWindowEvent();
    OnClosing();
    OnClosingEvent();
    LogEvent(event);
}

void RWindow::processOnOpen()
{
    auto event = RWindowEvent();
    OnOpen();
    OnOpenEvent();
    LogEvent(event);
}

#pragma endregion

std::string RWindow::GetTitle() const {
    return this->title;
}

/*
Vector2 RWindow::GetSize() const
{
    return {this->width, this->height};
}
*/

int RWindow::GetWidth() const
{
    return this->width;
}

int RWindow::GetHeight() const
{
    return this->height;
}

void RWindow::SetSizeWithoutEvent(const Vector2& size) {
    width = size.x;
    height = size.y;
}

void RWindow::SetLastKnownWindowSize(const Vector2& size) {
    lastKnownWindowSize = size;
}

void RWindow::SetRenderer(RenderingAPI api) {
    renderer = api;
}

void RWindow::SetSize(const Vector2& size) {
    this->width = size.x;
    this->height = size.y;
    this->SetSize(size.x, size.y);
}

bool RWindow::IsKeyDown(Key key) const {
    if (currentKeyboard.PressedKeys.contains(key))
        return currentKeyboard.PressedKeys.at(key);

    return false; // NOTE: Key may not be mapped!!
}

bool RWindow::IsMouseButtonDown(const MouseButton &button) const {
    // TODO: Implement MouseButton map
    return currentMouse.IsDown(button);
}


void RWindow::ManagedRefresh()
{
     auto begin = GetTimestamp();
     Refresh();
     auto end = GetTimestamp();

     float dt = ComputeElapsedFrameTimeSeconds(begin, end);

     UpdateFrameTiming(dt);
}

void RWindow::Refresh() {
    PollEvents();
    OnRefresh(delta_time);

    // Only call once and cache the result.
    currentMouse.Position = GetAccurateMouseCoordinates();

    /// TODO: Implement optional minimum epsilon to trigger a Mouse Update.
    if (currentMouse.Position != previousMouse.Position) {
        processMouseMove(previousMouse.Position, currentMouse.Position);
        previousMouse.Position = currentMouse.Position;
    }
}

float RWindow::ComputeElapsedFrameTimeSeconds(std::chrono::steady_clock::time_point start, std::chrono::steady_clock::time_point end) {
    auto frame_time = end - start;
    unsigned long int frame_time_us = std::chrono::duration_cast<std::chrono::microseconds>(frame_time).count();
    float frame_time_s = frame_time_us / (1000.f * 1000.f);
    return frame_time_s;
}

 std::chrono::steady_clock::time_point RWindow::GetTimestamp() {
    return std::chrono::steady_clock::now();
}

void RWindow::UpdateFrameTiming(float frame_time) {
    delta_time = frame_time;
    refresh_rate = 1.f / delta_time;
    refresh_rate_prev_5 = refresh_rate_prev_4;
    refresh_rate_prev_4 = refresh_rate_prev_3;
    refresh_rate_prev_3 = refresh_rate_prev_2;
    refresh_rate_prev_2 = refresh_rate_prev_1;
    refresh_rate_prev_1 = refresh_rate;
    avg_refresh_rate = (refresh_rate_prev_1 + refresh_rate_prev_2 + refresh_rate_prev_3 + refresh_rate_prev_4 + refresh_rate_prev_5) / 5.f;
    refresh_count++;
}


void RWindow::processMouseWheel(int scrolls)
{
    currentMouse.Wheel += scrolls;
    auto ev = MouseWheelEvent(scrolls);
    OnMouseWheel(ev);
    OnMouseWheelEvent(ev);

    previousMouse.Wheel = currentMouse.Wheel;
}


bool RWindow::IsResizable() const {
    return resizable;
}

bool RWindow::IsFullscreen() const {
    return fullscreen_mode;
}

bool RWindow::IsFocused() const {
    return focused;
}

 bool RWindow::IsVsyncEnabled() const {
     return vsync;
 }

 float RWindow::GetDeltaTime() const { return delta_time; }

 float RWindow::GetRefreshRate() const { return refresh_rate; }

 float RWindow::GetRefreshCounter() const { return refresh_count; }

 void RWindow::Close() {
     closing = true;
     processOnClose();
 }

 void RWindow::ForceClose() {
     Close();
     DestroyOSWindowHandle();
 }

 RWindow * RWindow::GetExtant() {
     return extant;
 }



 void RWindow::ForceCloseAndTerminateProgram() {
     ForceClose();
     exit(0);
 }


bool Input::IsKeyDown(const Key &key) {
    return extant->IsKeyDown(key);
}
bool Input::IsMouseButtonDown(const MouseButton &button) {
    return extant->IsMouseButtonDown(button);
}
Vector2 Input::GetMousePosition() {
    return extant->GetMouseCoordinates();
}
Vector2 Input::GetWindowSize() {
    return extant->GetSize();
}



